#!/usr/bin/python3

import datetime
import re
import jenkins
import json
import os
import signal

# import pprint
import requests
import sys
import yaml

signal.signal(signal.SIGINT, lambda signum, frame: sys.exit(0))

JenkinsException = Exception
NotFoundException = Exception

cfg = yaml.safe_load(open(os.path.join(os.environ["HOME"], ".config", "jenkins.yml")))
url = cfg[cfg["current"]]["url"]
username = cfg[cfg["current"]]["username"]
job_name = cfg[cfg["current"]]["job_name"]
ssh_private_key = cfg[cfg["current"]].get("ssh_private_key")
ssh_user = cfg[cfg["current"]].get("ssh_user")
token = cfg[cfg["current"]].get("token")

kwargs = {"username": username}
if token:
    kwargs["password"] = token
server = jenkins.Jenkins(url, **kwargs)
job = server.get_job_info(job_name, depth=0, fetch_all_builds=False)
# yaml.safe_dump(job, open("junk.yml", "w"))
task_nums = [build["number"] for build in job["builds"]]
# key is task num - val is full content of task
tasks_info = {}
# max task age in seconds
MAX_TASK_AGE = datetime.timedelta(seconds=int(os.environ.get("MAX_TASK_AGE", "86400")))
now = datetime.datetime.now()
now_tz = now.astimezone()

BUILD_ARTIFACT = "%(folder_url)sjob/%(short_name)s/%(number)s/artifact/%(artifact)s"

CONSOLE_TZ = "Z"
CONSOLE_TS_FMT = "%Y-%m-%d %H:%M:%S%z"


def time_to_local(ts_str, start_ts):
    """Convert console time string to local time."""
    dt_ts_str = start_ts.date().isoformat() + " " + ts_str + CONSOLE_TZ
    dt = datetime.datetime.strptime(dt_ts_str, CONSOLE_TS_FMT).astimezone(now_tz.tzinfo)
    return dt


def get_build_artifact(server, name, number, artifact):
    """Get artifacts from job
    :param name: Job name, ``str``
    :param number: Build number, ``str`` (also accepts ``int``)
    :param artifact: Artifact relative path, ``str``
    :returns: artifact to download, ``dict``
    """
    folder_url, short_name = server._get_job_folder(name)
    try:
        response = server.jenkins_open(
            requests.Request("GET", server._build_url(BUILD_ARTIFACT, locals()))
        )
        if response:
            return json.loads(response)
        else:
            raise JenkinsException("job[%s] number[%s] does not exist" % (name, number))
    except requests.exceptions.HTTPError:
        raise JenkinsException("job[%s] number[%s] does not exist" % (name, number))
    except ValueError:
        raise JenkinsException(
            "Could not parse JSON info for job[%s] number[%s]" % (name, number)
        )
    except NotFoundException:
        # This can happen if the artifact is not found
        return None


def get_pr_status_label(task, short=True):
    for action in task["actions"]:
        if action["_class"] == "hudson.model.ParametersAction":
            for param in action["parameters"]:
                if param["name"] == "pipeline_state_reporter_options":
                    label = param["value"].split("=")[1]
                    if label.endswith("/(citool)"):
                        label = label.replace("/(citool)", "")
                    if short:
                        match = re.match(r"^(RHEL-\d+[.]\d+)[^/]*(/.+)$", label)
                        if match:
                            label = match.group(1) + match.group(2)
                        else:
                            match = re.match(r"^(CentOS-Stream-\d+)(/.+)$", label)
                            if match:
                                label = match.group(1).replace(
                                    "-Stream", ""
                                ) + match.group(2)
                            else:
                                match = re.match(r"^(CentOS-\d+)[^/]+(/.+)$", label)
                                if match:
                                    label = match.group(1) + match.group(2)
                    return label


def get_pr_info(task):
    for action in task["actions"]:
        if action["_class"] == "hudson.model.ParametersAction":
            for param in action["parameters"]:
                if param["name"] == "github_options":
                    val = param["value"].split("=")[1]
                    org, repo, pr = val.split(":")[0:3]
                    if org == "linux-system-roles":
                        return (repo, pr)
                    else:
                        return (org + "/" + repo, pr)


def get_queued_time(task):
    for action in task["actions"]:
        if action.get("_class") == "jenkins.metrics.impl.TimeInQueueAction":
            return str(int(action["buildableDurationMillis"] / 1000))


def get_test_status(task):
    status = ""
    for action in task["actions"]:
        if "_class" not in action:
            continue
        if action["_class"] == "com.jenkinsci.plugins.badge.action.BadgeAction":
            if "pipeline" in action["text"]:
                continue
            status = action["text"]
            break
    if not status:
        status = task["result"]
    if status == "Tests did not run correctly":
        status = "CANCELLED"
    status = status.replace("Tests ", "")
    return status


def format_queued_task(task, ignored):
    label = get_pr_status_label(task)
    role, prnum = get_pr_info(task)
    ts = datetime.datetime.fromtimestamp(task["inQueueSince"] / 1000).isoformat(
        timespec="seconds"
    )
    if task["why"].startswith("Waiting for next available executor"):
        why = "waiting on executor"
    else:
        why = task["why"]
    return (str(task["id"]), ts, role, prnum, label, why)


def format_running_task(task, ts):
    label = get_pr_status_label(task)
    role, prnum = get_pr_info(task)
    queue_time = get_queued_time(task)
    tsstr = ts.isoformat(timespec="seconds")
    return (task["id"], tsstr, role, prnum, label, queue_time)


def format_completed_task(task, ts):
    label = get_pr_status_label(task)
    role, prnum = get_pr_info(task)
    queue_time = get_queued_time(task)
    status = get_test_status(task)
    duration = str(int(task["duration"] / 1000))
    tsstr = ts.isoformat(timespec="seconds")
    return (task["id"], tsstr, role, prnum, label, duration, queue_time, status)


FORMATS = {
    "queued": {
        "hdr": ("QueueID", "Queued Since", "Role", "PR", "Platform", "Queue Reason"),
        "fmt": "{:8s} {:19s} {:15s} {:3s} {:22s} {:20s}",
        "fn": format_queued_task,
    },
    "running": {
        "hdr": ("TaskID", "Started At", "Role", "PR", "Platform", "Queue Time"),
        "fmt": "{:8s} {:19s} {:15s} {:3s} {:22s} {:10s}",
        "fn": format_running_task,
    },
    "completed": {
        "hdr": (
            "TaskID",
            "Started At",
            "Role",
            "PR",
            "Platform",
            "Duration",
            "Queue Time",
            "Status",
        ),
        "fmt": "{:8s} {:19s} {:15s} {:3s} {:22s} {:8s} {:10s} {:10s}",
        "fn": format_completed_task,
    },
}


def format_fields(task, task_state, ts, is_header=False):
    # task_state is one of queued, running, completed
    fmt = FORMATS[task_state]["fmt"]
    if is_header:
        data = FORMATS[task_state]["hdr"]
    else:
        data = FORMATS[task_state]["fn"](task, ts)
    # print(f"fmt {fmt} data {data}")
    return fmt.format(*data)


def get_task_ts(server, task_num):
    """Return the task info and timestamp."""
    global tasks_info
    task, ts = tasks_info.get(task_num, (None, None))
    if task is None:
        task = server.get_build_info(job_name, task_num)
        ts = datetime.datetime.fromtimestamp(task["timestamp"] / 1000)
        tasks_info[task_num] = (task, ts)
    return (task, ts)


def task_iter(task_nums, server):
    """Iterate through task nums returning task info and timestamp."""
    for num in task_nums:
        task, ts = get_task_ts(server, num)
        if now - ts < MAX_TASK_AGE:
            yield (task, ts)
        else:
            break


def print_running_tasks(server, task_nums, args):
    print(format_fields(None, "running", None, True))
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            print(format_fields(task, "running", ts))
    # pprint.pprint(lastbuild)
    # relpath = "work-tests_configure_ha_cluster.ymlT7FqqE/ansible-output.txt"
    # #not in version 1.7.0
    # artifact = get_build_artifact(server, job_name, lastnum, relpath)
    # pprint.pprint(artifact)
    # console = server.get_build_console_output(job_name, lastnum)
    # console_lines = console.split("\n")
    # pprint.pprint(console_lines[-10:])
    # not permitted
    # plugins = server.get_plugins()
    # print(plugins)
    # node = server.get_node_info(lastbuild["builtOn"])
    # pprint.pprint(node)


def print_queued_tasks(server, task_nums, args):
    print(format_fields(None, "queued", None, True))
    queue_info = server.get_queue_info()
    size = 0
    for task in queue_info:
        size = size + 1
        if task["task"]["name"] == job_name:
            print(format_fields(task, "queued", None))
    print(f"Queue size: {size}")


def print_completed_tasks(server, task_nums, args):
    print(format_fields(None, "completed", None, True))
    for task, ts in task_iter(task_nums, server):
        if not task["result"] is None:
            print(format_fields(task, "completed", ts))


def cancel_tasks(server, task_nums, args):
    """Cancel given tasks."""
    for num in args:
        print(f"Stopping {num}")
        server.stop_build(job_name, num)


def print_task_info(server, task_nums, args):
    """Print info for given build numbers."""
    for num in args:
        task = server.get_build_info(job_name, int(num))
        yaml.safe_dump(task, sys.stdout)


def get_node_info_for_task(server, task_num):
    """Print the node info for the given task."""
    task = server.get_build_info(job_name, task_num)
    node_name = task["builtOn"]
    node = server.get_node_info(node_name)
    description = node["description"]
    match = re.search(r"HOSTNAME=([0-9.]+)", description)
    if match:
        return {"node_name": node_name, "ip": match.group(1)}
    else:
        print(f"ERROR: for node_name {node_name} description unknown {description}")
        return {"node_name": node_name, "ip": "unknown"}


def get_task_tests_info(server, task_num):
    """Get the info for all of the tests of a task given a task num."""
    pat_pr = r"^ +lsr-github:github --pull-request= --pull-request=linux-system-roles:(?P<role>[^:]+):(?P<pr>[^:]+):.*$"
    rx_pr = re.compile(pat_pr)
    pat_test = (
        r"^[|] (?:dist-)?git-[^/]+/tests/(?P<test>[^ ]+) +[|] +(?P<stage>[^ ]+) +[|]"
        r" +(?P<state>[^ ]+) +[|] +(?P<result>[^ ]+) +[|] +(?P<arch>[^ ]+) +(?P<platform>[^ ]+) .*"
    )
    rx_test = re.compile(pat_test)
    pat_guest_id = r" [|] +([a-zA-Z0-9]{8}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{4}-[a-zA-Z0-9]{12}) +[|]"
    rx_guest_id = re.compile(pat_guest_id)
    pat_workspace = r"^Building remotely on .* in workspace (?P<workspace>[^ ]+)$"
    rx_workspace = re.compile(pat_workspace)
    pat_work_test_dir = (
        r"^.*(?:dist-)?git-[^ ]+ working directory 'work-([^.]+.yml)([^']+)'.*$"
    )
    rx_work_test_dir = re.compile(pat_work_test_dir)
    pat_provision_start = (
        r"^.*\[(?P<ts_prov>[0-9:]+)\] .*\[(?:dist-)?git-[^/]+/tests/(?P<test>tests[^.]+[.]yml)\].* "
        r"starting guest provisioning.*$"
    )
    rx_provision_start = re.compile(pat_provision_start)
    pat_guest_setup_start = (
        r"^.*\[(?P<ts_setup>[0-9:]+)\] .*\[(?:dist-)?git-[^/]+/tests/(?P<test>tests[^.]+[.]yml)\].* "
        r"starting guest setup.*$"
    )
    rx_guest_setup_start = re.compile(pat_guest_setup_start)
    pat_test_start = (
        r"^.*\[(?P<ts_start>[0-9:]+)\] .*\[(?:dist-)?git-[^/]+/tests/(?P<test>tests[^.]+[.]yml)\].* "
        r"starting tests execution.*$"
    )
    rx_test_start = re.compile(pat_test_start)
    pat_test_end = (
        r"^.*\[(?P<ts_end>[0-9:]+)\] .*\[(?:dist-)?git-[^/]+/tests/(?P<test>tests[^.]+[.]yml)\].* "
        r"cleanup finished.*$"
    )
    rx_test_end = re.compile(pat_test_end)

    task, ts = get_task_ts(server, task_num)
    console = server.get_build_console_output(job_name, task_num)
    console_lines = console.split("\n")
    tests = {}
    info = {"tests": tests, "start": ts.isoformat(timespec="seconds")}
    last_test = ""
    for line in console_lines:
        match = rx_guest_id.search(line)
        if match and last_test:
            tests[last_test]["guest_id"] = match.group(1)
        elif last_test:
            tests[last_test]["guest_id"] = "unknown"
        match = rx_test.match(line)
        if match:
            test = match.group("test")
            tests.setdefault(test, {}).update(match.groupdict())
            info.update(match.groupdict())
            last_test = test
        else:
            last_test = ""
        match = rx_pr.match(line)
        if match:
            info.update(match.groupdict())
        match = rx_workspace.match(line)
        if match:
            info.update(match.groupdict())
        match = rx_work_test_dir.match(line)
        if match:
            test = match.group(1)
            work_dir = "work-" + test + match.group(2)
            tests.setdefault(test, {})["work_dir"] = work_dir
        for rx in (
            rx_provision_start,
            rx_guest_setup_start,
            rx_test_start,
            rx_test_end,
        ):
            match = rx.match(line)
            if match:
                test = match.group(2)
                grp_dict = match.groupdict()
                for kk, vv in match.groupdict().items():
                    if kk == "test":
                        continue
                    dt = time_to_local(vv, ts)
                    grp_dict[kk] = dt.strftime("%Y-%m-%dT%H:%M:%S")
                tests.setdefault(test, {}).update(grp_dict)

    node_info = get_node_info_for_task(server, task_num)
    info.update(node_info)
    return info


def print_task_tests_info(server, task_nums, args):
    """Print tests information for a given task."""
    info = get_task_tests_info(server, int(args[0]))
    print(
        f"Role:{info['role']} PR:{info['pr']} Platform:{info['platform']} Arch:{info['arch']} Start:{info['start']}"
    )
    print(f"Node:{info['node_name']} IP:{info['ip']} Workspace:{info['workspace']}")
    fmt = "{:20s} {:10s} {:10s} {:36s} {} {}"
    print(fmt.format("Stage", "State", "Result", "GuestID", "Test", "Workdir"))
    for test, data in info["tests"].items():
        print(
            fmt.format(
                data["stage"],
                data["state"],
                data["result"],
                data["guest_id"],
                test,
                data.get("work_dir", "unknown"),
            )
        )
        print(
            f"ts_prov:{data.get('ts_prov', 'None')} ts_setup:{data.get('ts_setup', 'None')} "
            f"ts_start:{data.get('ts_start', 'None')} ts_end:{data.get('ts_end', 'None')}"
        )


def print_task_console(server, task_nums, args):
    console = server.get_build_console_output(job_name, int(args[0]))
    last = int(args[1])
    console_lines = "\n".join(console.split("\n")[-last:])
    print(console_lines)


def print_node_info(server, task_nums, args):
    """Print info for given build node."""
    for node_name in args:
        node = server.get_node_info(node_name)
        yaml.safe_dump(node, sys.stdout)


# def get_workspace_file(ip, workspace_dir, work_test_dir, guest_id, file_type):
#     """Get the file from the workspace dir on the Jenkins server."""
#     if file_type == "ansible":
#         pass
#     cmd = f"scp -i {ssh_private_key} {ssh_user}@{ip}:"
#     subprocess.run(cmd)


def cancel_duplicate_tasks(server, task_nums, args):
    """Stop duplicate queued and running tasks.  Keep the newest one."""
    found_keys = set()
    queue_info = server.get_queue_info()
    for task in queue_info:
        if task["task"]["name"] == job_name:
            label = get_pr_status_label(task, False)
            role, prnum = get_pr_info(task)
            key = f"{label}:{role}:{prnum}"
            if key in found_keys:
                tsstr = datetime.datetime.fromtimestamp(
                    task["inQueueSince"] / 1000
                ).isoformat(timespec="seconds")
                print(f"Cancelling duplicate queued task {task['id']} {key} {tsstr}")
                server.cancel_queue(task["id"])
            else:
                found_keys.add(key)
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            label = get_pr_status_label(task, False)
            role, prnum = get_pr_info(task)
            key = f"{label}:{role}:{prnum}"
            if key in found_keys:
                tsstr = ts.isoformat(timespec="seconds")
                print(f"Cancelling duplicate running task {task['id']} {key} {tsstr}")
                server.stop_build(job_name, task["id"])
            else:
                found_keys.add(key)


def cancel_pr_tasks(server, task_nums, args):
    """Cancel tasks for given roles and prs."""
    prs = set(args)
    queue_info = server.get_queue_info()
    for task in queue_info:
        if task["task"]["name"] == job_name:
            role, prnum = get_pr_info(task)
            key = f"{role}:{prnum}"
            if key in prs:
                tsstr = datetime.datetime.fromtimestamp(
                    task["inQueueSince"] / 1000
                ).isoformat(timespec="seconds")
                print(f"Cancelling pr {key} queued task {task['id']} {key} {tsstr}")
                server.cancel_queue(task["id"])
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            role, prnum = get_pr_info(task)
            key = f"{role}:{prnum}"
            if key in prs:
                tsstr = ts.isoformat(timespec="seconds")
                print(f"Cancelling pr {key} running task {task['id']} {key} {tsstr}")
                server.stop_build(job_name, task["id"])


def print_test_failures(server, task_nums, args):
    """Print test failures for tasks matching args."""
    role, prnum, label = None, None, None
    if len(args) > 0:
        role = args[0]
    if len(args) > 1:
        prnum = args[1]
    if len(args) > 2:
        label = args[2]
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            task_label = get_pr_status_label(task)
            task_role, task_prnum = get_pr_info(task)
            if (
                role == task_role
                and (prnum is None or prnum == task_prnum)
                and (label is None or label == task_label)
            ):
                info = get_task_tests_info(server, int(task["id"]))
                for test, data in info["tests"].items():
                    if data["result"] not in ["PASSED", "UNDEFINED"]:
                        print(
                            f"{task_role} {task_prnum} {task_label} {test} {info['start']}"
                        )


def print_test_results(server, task_nums, args):
    """Print test failures for tasks matching args."""
    role, prnum, test_pattern = None, None, None
    if len(args) > 0:
        role = args[0]
    if len(args) > 1:
        prnum = args[1]
    if len(args) > 2:
        test_pattern = re.compile(args[2])
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            task_label = get_pr_status_label(task)
            task_role, task_prnum = get_pr_info(task)
            if role == task_role and (prnum is None or prnum == task_prnum):
                info = get_task_tests_info(server, int(task["id"]))
                for test, data in info["tests"].items():
                    if test_pattern.search(test) is not None:
                        print(
                            f"{task_role} {task_prnum} {task_label} {test} {data['result']} {info['start']}"
                        )


def print_pr_tests_info(server, task_nums, args):
    """Print test failures for tasks matching args."""
    role, prnum, test_pattern, result = None, None, None, None
    if len(args) > 0:
        role = args[0]
    if len(args) > 1:
        prnum = args[1]
    if len(args) > 2:
        test_pattern = re.compile(args[2])
    if len(args) > 3:
        result = args[3]
    for task, ts in task_iter(task_nums, server):
        if task["result"] is None:
            task_label = get_pr_status_label(task)
            task_role, task_prnum = get_pr_info(task)
            if role == task_role and (prnum is None or prnum == task_prnum):
                info = get_task_tests_info(server, int(task["id"]))
                for test, data in info["tests"].items():
                    if test_pattern.search(test) is not None and (
                        result is None or result == data["result"]
                    ):
                        print(
                            f"{task_role} {task_prnum} {task_label} {test} {data['result']} {info['start']}"
                        )


if len(sys.argv) > 1:
    locals()[sys.argv[1]](server, task_nums, sys.argv[2:])
else:
    print("Queued tasks:")
    print_queued_tasks(server, task_nums, [])
    print("\nRunning tasks:")
    print_running_tasks(server, task_nums, [])
    print("\nCompleted tasks:")
    print_completed_tasks(server, task_nums, [])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright: (c) 2020, Red Hat, Inc.
# SPDX-License-Identifier: MIT
"""This script is used to:
* Check all of the roles for new versions, and update `collection_release.yml`
  with the updated `ref` fields.
* Convert the roles to collection using [lsr_role2collection.py](#lsr_role2collectionpy)
* Update the version in `galaxy.yml`
* Build the collection file using `ansible-galaxy collection build`
* Check the collection using `galaxy-importer`
* Publish the collection using `ansible-galaxy collection publish`
"""

import argparse
import logging
import os
import re
import shutil
import subprocess
import sys
import tempfile
from datetime import datetime

try:
    import yaml
except ImportError:
    import ruamel.yaml as yaml

import json

from pkg_resources import parse_version

DEFAULT_GIT_SITE = "https://github.com"
DEFAULT_GIT_ORG = "linux-system-roles"


def run_cmd(cmdlist, cwd=None, env=None):
    """Run the given cmdlist using subprocess.  Debug log the output.
    If check is true, this function will work like check_call.
    The return value is like subprocess.run"""
    kwargs = {"env": os.environ}
    if env:
        kwargs["env"].update(env)
    rc = subprocess.run(
        cmdlist,
        cwd=cwd,
        encoding="utf-8",
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )
    logging.debug(
        f"{' '.join(cmdlist)} returned {rc.returncode} stdout {rc.stdout} stderr {rc.stderr}"
    )
    if rc.returncode != 0:
        raise subprocess.CalledProcessError(
            rc.returncode, cmdlist, rc.stdout, rc.stderr
        )
    return rc


def lsr_parse_version(v_str):
    test_v = parse_version("1")  # guaranteed to work
    if v_str is None:
        v = parse_version("0.0.0")
    else:
        v = parse_version(v_str)
    if not isinstance(v, test_v.__class__):
        raise ValueError(f"Error: {v_str} is not a proper version number")
    if not hasattr(v, "release"):
        if hasattr(v, "_version"):
            setattr(v, "release", v._version.release)
    return v


def check_versions_updated(cur_ref, new_ref, versions_updated):
    """Compare versions and update list.

    If cur_ref and new_ref are both valid semantic versions,
    compare them, and indicate in versions_updated if the major,
    minor, and/or micro versions were updated.  One exception is
    when the major version is upgraded from 0 to 1 - in that case,
    do not mark this as requiring a major version change.
    Another special case is adding a new role - in that case, cur_ref
    is None - we do not want to bump the major version, only the
    minor version."""

    if cur_ref is None:
        versions_updated[1] = True
        return
    try:
        cur_v = lsr_parse_version(cur_ref)
        new_v = lsr_parse_version(new_ref)
        for idx in range(0, 3):
            if cur_v.release[idx] != new_v.release[idx]:
                if idx > 0 or cur_v.release[idx] > 0:
                    versions_updated[idx] = True
    except ValueError as exc:
        logging.debug(f"Could not compare version {cur_ref} to {new_ref}: {exc}")
        if cur_ref != new_ref:
            versions_updated[3] = True


def comp_versions(cur_ref, new_ref):
    """Compare versions.

    Return values:
    -1 - if cur_ref < new_ref
     0 - if cur_ref == new_ref
     1 - if cur_ref > new_ref"""

    try:
        v_cur_ref = lsr_parse_version(cur_ref)
        v_new_ref = lsr_parse_version(new_ref)
        if v_cur_ref < v_new_ref:
            return -1
        elif v_cur_ref == v_new_ref:
            return 0
        else:
            return 1
    except ValueError as exc:
        logging.debug(f"Could not compare version {cur_ref} to {new_ref}: {exc}")
        if cur_ref == new_ref:
            return 0
        else:
            return -1  # assume new is "newer" than cur


def get_latest_tag_hash(args, rolename, cur_ref, org, repo, use_commit_hash):
    """
    Get the latest tag, hash, and tag_is_latest from the upstream repo.

    Clone and/or update the local copies of the repos.  Get
    the latest tag and/or commit hash for each repo.  Indicate if
    the tag is the latest commit.
    """
    roledir = os.path.join(args.src_path, rolename)
    # clone and/or update role repo
    if os.path.isdir(roledir):
        _ = run_cmd(["git", "fetch"], roledir)
    else:
        _ = run_cmd(
            [
                "git",
                "-c",
                "advice.detachedHead=false",
                "clone",
                "-q",
                f"{DEFAULT_GIT_SITE}/{org}/{repo}",
                roledir,
            ],
        )
    branch_output = run_cmd(["git", "branch", "-r"], roledir)
    # determine what is the main branch, check it out, and update it
    mmatch = re.search(r"origin/HEAD -> origin/(\w+)", branch_output.stdout)
    main_branch = mmatch.group(1)
    _ = run_cmd(["bash", "-c", f"git checkout {main_branch}; git pull --tags"], roledir)
    commit_msgs = ""
    # see if there have been any commits since the last time we checked
    if cur_ref:
        count_output = run_cmd(
            ["bash", "-c", f"git log --oneline {cur_ref}.. | wc -l"],
            roledir,
        )
    else:
        count_output = run_cmd(
            ["bash", "-c", "git log --oneline | wc -l"],
            roledir,
        )
    tag, commit_hash, n_commits, prev_tag = None, None, "0", None
    if count_output.stdout == "0":
        logging.debug(f"no changes to role {rolename} since ref {cur_ref}")
    else:
        # get latest tag and commit hash
        describe_cmd = ["git", "describe", "--tags", "--long", "--abbrev=40"]
        describe_output = run_cmd(describe_cmd, roledir)
        tag, n_commits, g_hash = describe_output.stdout.strip().rsplit("-", 2)
        # commit hash - skip leading "g"
        commit_hash = g_hash[1:]
        if n_commits != "0" and use_commit_hash:
            # get commit messages to use for changelog
            log_cmd = [
                "git",
                "log",
                "--oneline",
                "--no-merges",
                "--reverse",
                "--pretty=format:- %s",
            ]
            if cur_ref:
                log_cmd.append(f"{cur_ref}..")
            log_output = run_cmd(log_cmd, roledir)
            commit_msgs = log_output.stdout.replace("\\r", "")
        # get previous tag in case cur_ref is a commit hash
        if cur_ref:
            try:
                describe_cmd = [
                    "git",
                    "describe",
                    "--tags",
                    "--long",
                    "--abbrev=40",
                    cur_ref,
                ]
                describe_output = run_cmd(describe_cmd, roledir)
                prev_tag = describe_output.stdout.strip().split("-")[0]
                if prev_tag == cur_ref:
                    prev_tag = None  # cur_ref was already a valid tag
            except subprocess.CalledProcessError:
                prev_tag = None  # no previous tag
        else:
            prev_tag = None  # no previous tag
    if args.no_update and cur_ref:
        # make sure cur_ref is checked out
        _ = run_cmd(["git", "checkout", cur_ref], roledir)
    return (tag, commit_hash, n_commits == "0", commit_msgs, prev_tag)


def process_ignore_and_lint_files(args, coll_dir):
    """Create collection ignore-VER.txt and .ansible-lint files from roles."""
    ignore_file_dir = os.path.join(coll_dir, "tests", "sanity")
    # role2collection is currently done using ruamel rt with a long line length,
    # because we can't ensure converted lines will be wrapped preserving
    # Jinja and ansible syntax.  This means we trigger a lot of warnings like
    # this when importing into Galaxy:
    # roles/file.yml:61: yaml line too long (187 > 160 characters)
    # the only thing we can do is suppress these
    ansible_lint = {"skip_list": ["yaml[line-length]"]}
    for role_name in os.listdir(args.src_path):
        roledir = os.path.join(args.src_path, role_name)
        if (
            os.path.isdir(roledir)
            and os.path.isfile(os.path.join(roledir, "tasks", "main.yml"))
            and role_name in args.include
        ):
            for file_name in os.listdir(roledir):
                if file_name.startswith(".sanity-ansible-ignore-"):
                    if not os.path.isdir(ignore_file_dir):
                        os.mkdir(ignore_file_dir)
                    match = re.match(
                        r"^[.]sanity-ansible-ignore-(\d[\d.]+)[.]txt$", file_name
                    )
                    if match and match.groups() and match.group(1):
                        ver = match.group(1)
                        ignore_file = os.path.join(ignore_file_dir, f"ignore-{ver}.txt")
                        role_ignore_file = os.path.join(roledir, file_name)
                        with open(ignore_file, "a") as ign_fd:
                            with open(role_ignore_file, "r") as role_ign_fd:
                                ign_fd.write(role_ign_fd.read())
                elif file_name == ".ansible-lint":
                    role_ansible_lint = yaml.safe_load(
                        open(os.path.join(roledir, file_name))
                    )
                    # ensure we suppress line-length checks
                    for key, items in role_ansible_lint.items():
                        if key not in ansible_lint:
                            ansible_lint[key] = role_ansible_lint[key]
                            continue
                        for item in items:
                            if item not in ansible_lint[key]:
                                if isinstance(ansible_lint[key], list):
                                    ansible_lint[key].append(item)
                                elif isinstance(ansible_lint[key], dict):
                                    ansible_lint[key][item] = items[item]
    yaml.safe_dump(ansible_lint, open(os.path.join(coll_dir, ".ansible-lint"), "w"))


def get_role_changelog(args, rolename, cur_ref, new_ref, commit_msgs):
    """
    Retrieve the matched changelogs from CHANGELOG.md and
    return them as a string.
    """
    _changelog = ""
    if comp_versions(cur_ref, new_ref) >= 0:
        return _changelog
    _changelog = "### {}\n".format(rolename)
    if commit_msgs:
        # make a fake changelog for compact_coll_changelog
        _changelog = "{}\n##### Bug Fixes\n\n{}".format(_changelog, commit_msgs)
        return _changelog
    elif cur_ref is None:
        # make a fake changelog for compact_coll_changelog for new role
        _changelog = "{}\n##### New Features\n\n- New Role".format(_changelog)
        return _changelog
    _changelogmd = os.path.join(args.src_path, rolename, "CHANGELOG.md")
    if not os.path.exists(_changelogmd):
        return ""
    _print = False
    with open(_changelogmd, "r") as cl_fd:
        _prev = ""
        for _cl in cl_fd:
            cl = _cl.rstrip()
            if cl.startswith("[{}]".format(new_ref)):
                _print = True
                _changelog = "{}\n#### {}".format(_changelog, cl)
            elif cl.startswith("[{}]".format(cur_ref)):
                break
            elif _print:
                if cl.lower() == "- none":
                    if _prev and _changelog.endswith(_prev + "\n"):
                        _changelog = _changelog[: -len(_prev + "\n")]
                    _prev = "none"
                elif cl.startswith("["):
                    _changelog = "{}\n#### {}".format(_changelog, cl)
                elif not cl.startswith("----"):
                    _matched = False
                    for _title in args.changelog_sections:
                        if cl.lower() == _title.lower():
                            _prev = "\n##{}".format(_title)
                            _changelog = "{}{}".format(_changelog, _prev)
                            _matched = True
                            break
                    if not _matched:
                        if _prev == "none":
                            _prev = ""
                        else:
                            _changelog = "{}\n{}".format(_changelog, cl)

    logging.info("get_role_changelog - returning\n%s", _changelog)
    return _changelog


def role_to_collection(
    args,
    rolename,
    namespace,
    collection_name,
    collection_readme,
):
    """Convert the role to collection format."""
    subrole_prefix = f"private_{rolename}_subrole_"
    cmd = [
        sys.executable,
        "lsr_role2collection.py",
        "--src-owner",
        args.src_owner,
        "--role",
        rolename,
        "--src-path",
        args.src_path,
        "--dest-path",
        args.dest_path,
        "--namespace",
        namespace,
        "--collection",
        collection_name,
        "--subrole-prefix",
        subrole_prefix,
        "--readme",
        collection_readme,
    ]
    extra_mapping = ""
    comma = ""
    if rolename == "sshd":
        # HACK - special case for ansible-sshd - not fully qualified
        extra_mapping = f"ansible-sshd:{namespace}.{collection_name}.{rolename}"
        comma = ","
    if args.extra_mapping:
        extra_mapping = extra_mapping + comma + args.extra_mapping
    if extra_mapping:
        cmd.extend(["--extra-mapping", extra_mapping])
    _ = run_cmd(cmd)


def update_galaxy_version(args, galaxy, versions_updated):
    if args.no_update:
        return
    if (
        not versions_updated[3]
        and any(versions_updated[0:3])
        and not args.no_auto_version
    ):
        galaxy_ver = lsr_parse_version(galaxy["version"])
        major, minor, micro = galaxy_ver.release
        if versions_updated[0]:
            major = major + 1
            minor = 0
            micro = 0
        elif versions_updated[1]:
            minor = minor + 1
            micro = 0
        elif versions_updated[2]:
            micro = micro + 1
        galaxy["version"] = f"{major}.{minor}.{micro}"
    elif args.new_version:
        galaxy["version"] = args.new_version
    else:
        raise Exception(
            "No new-version given and cannot auto-update version.  "
            "Please examine the changes to the collection, determine what "
            "the new semantic version will be, and re-run this script with "
            "the --new-version argument"
        )
    # not worth trying to figure out how to do a round-trip yaml write here
    # not work using ruamel here
    # this is simple enough for re.sub
    with open(args.galaxy_yml.name) as gf:
        galaxy_str = gf.read()
        galaxy_str = re.sub(
            r"^version:.*$", f'version: "{galaxy["version"]}"', galaxy_str, flags=re.M
        )
    with open(args.galaxy_yml.name, "w") as gf:
        gf.write(galaxy_str)


def create_collection_extra_files(args, coll_dir, galaxy=None):
    collection_requirements = os.path.join(
        "lsr_role2collection", "collection_requirements.txt"
    )
    collection_requirements_dest = os.path.join(coll_dir, "requirements.txt")
    collection_bindep = os.path.join("lsr_role2collection", "collection_bindep.txt")
    collection_bindep_dest = os.path.join(coll_dir, "bindep.txt")
    collection_req_yml_dest = os.path.join(coll_dir, "requirements.yml")
    collection_aee_dest = os.path.join(coll_dir, "execution-environment.yml")
    collection_yamllint = os.path.join("lsr_role2collection", "yamllint.yml")
    collection_yamllint_dest = os.path.join(coll_dir, ".yamllint.yml")
    shutil.copy(collection_yamllint, collection_yamllint_dest)
    # If --rpm, files such as galaxy.yml in coll_dir are being used.
    process_ignore_and_lint_files(args, coll_dir)
    if not args.rpm:
        if galaxy:
            with open(os.path.join(coll_dir, "galaxy.yml"), "w") as galaxy_fd:
                yaml.safe_dump(galaxy, galaxy_fd)
        else:
            shutil.copy(args.galaxy_yml.name, coll_dir)
        shutil.copy(args.collection_release_yml.name, coll_dir)
        ee_dependencies = []
        if os.path.exists(collection_requirements):
            shutil.copy(collection_requirements, collection_requirements_dest)
            ee_dependencies.append("python: requirements.txt")
        if os.path.exists(collection_bindep):
            shutil.copy(collection_bindep, collection_bindep_dest)
            ee_dependencies.append("system: bindep.txt")
        if galaxy["dependencies"]:
            with open(collection_req_yml_dest, "w") as crf:
                crf.write("---\ncollections:\n")
                for k in galaxy["dependencies"].keys():
                    crf.write("  - name: {0}\n".format(k))
            ee_dependencies.append("galaxy: requirements.yml")
        with open(collection_aee_dest, "w") as eef:
            eef.write("---\nversion: 1\n\n")
            eef.write("build_arg_defaults:\n")
            eef.write("  EE_BASE_IMAGE: {0}\n\n".format(args.ee_base_image))
            eef.write("ansible_config: 'ansible.cfg'\n\n")
            if len(ee_dependencies):
                eef.write("dependencies:\n")
                for eed in ee_dependencies:
                    eef.write("  {0}\n".format(eed))


def build_collection(args, coll_dir, galaxy=None):
    create_collection_extra_files(args, coll_dir, galaxy)
    # removing dot files/dirs
    keep_files = set([".ansible-lint", ".yamllint.yml"])
    for file_name in os.listdir(coll_dir):
        if file_name in keep_files:
            continue
        if re.match(r"^[.][a-zA-Z]+", file_name):
            full_name = os.path.join(coll_dir, file_name)
            if os.path.isdir(full_name):
                shutil.rmtree(full_name)
            else:
                os.unlink(full_name)

    if shutil.which("ansible-galaxy"):
        build_args = ["ansible-galaxy", "collection", "build", "-v"]
        if args.force:
            build_args.append("-f")
        build_args.append(coll_dir)
        _ = run_cmd(build_args, args.dest_path)
    else:
        logging.info("ansible-galaxy is skipped since it is not available.")


NF = "New Features"
BF = "Bug Fixes"


def gather_changes(vstr, changes):
    output = ""
    if vstr:
        output = "{}\n---------------------".format(vstr)
    if changes[NF]:
        output = "{}\n### New Features\n\n".format(output)
        for c in changes[NF]:
            output = "{}{}".format(output, c)
    if changes[BF]:
        output = "{}\n### Bug Fixes\n\n".format(output)
        for c in changes[BF]:
            output = "{}{}".format(output, c)
    return output


def compact_coll_changelog(input_changelog):
    """ """
    output_changelog = ""
    enabled = False
    in_code = False
    vmatched = False
    changes = {}
    changes[NF] = []
    changes[BF] = []
    vstr = ""
    tag = ""
    role = ""
    for _cl in input_changelog.splitlines():
        cl = _cl.rstrip()
        if cl == "" or cl.startswith("#### [") or cl == "- none":
            continue
        if cl == "##### Other Changes":
            enabled = False
            continue
        if cl.startswith("```"):
            in_code = not in_code  # toggle
            continue
        _mobj = re.match(r"\[\d*[.]\d*[.]\d*\] - \d{4}-\d{2}-\d{2}", cl)
        if _mobj:
            vmatched = True
            enabled = False
            if vstr and (changes[NF] or changes[BF]):
                output_changelog = "{}\n{}".format(
                    output_changelog, gather_changes(vstr, changes)
                )
            changes[NF] = []
            changes[BF] = []
            vstr = _mobj[0]
            continue
        _mobj = re.match(r"-*", cl)
        if vmatched and _mobj[0]:
            vmatched = False
            enabled = True
        else:
            if cl == "##### New Features" or cl == "##### New features":
                tag = NF
            elif cl == "##### Bug Fixes" or cl == "##### Bug fixes":
                tag = BF
            else:
                # ### ROLENAME
                _mobj = re.match(r"### (.*)", cl)
                if _mobj and _mobj.group(1):
                    role = _mobj.group(1)
                    enabled = True
                elif enabled and not in_code:
                    # Retrieves itemized changes
                    _mobj = re.match(r"- (.*)", cl)
                    if _mobj and _mobj.group(1):
                        change = _mobj.group(1)
                        if tag and role:
                            changes[tag].append("- {0} - {1}\n".format(role, change))
    if changes[NF] or changes[BF]:
        output_changelog = "{}\n{}".format(
            output_changelog, gather_changes("", changes)
        )
    else:
        output_changelog = (
            "{}\n\n### Other Changes\n\n- no user-visible changes\n"
        ).format(output_changelog)
    return output_changelog


def conv_md2rest(changelog_md_path, changelog_rest_path):
    """
    Convert the md format to the reStructuredText format
    """
    rval = ""
    try:
        import pypandoc

        rval = pypandoc.convert_file(changelog_md_path, "rst", format="md")
    except (IOError, ImportError):
        raise Exception("Failed to import pypandoc module")
    with open(changelog_rest_path, "w", encoding="utf-8") as f:
        f.write(rval)
    # Verify the converted reST
    try:
        from rst2html import rst2html

        with open(changelog_rest_path, "r") as f:
            rst_text = f.read()
        html, warning = rst2html(rst_text, report_level=3)
        if len(warning) > 0:
            raise Exception(
                "Converted CHANGELOG.rst has a problem as html - {}".format(warning)
            )
    except (IOError, ImportError):
        logging.warning("Failed to import rst2html module")
        pass


def convert_md2rst(coll_dir):
    """
    Convert CHANGELOG.md to CHANGELOG.rst
    """
    coll_changelog_path = os.path.join(coll_dir, "docs", "CHANGELOG.md")
    if os.path.exists(coll_changelog_path):
        coll_changelog_rest_path = os.path.join(coll_dir, "CHANGELOG.rst")
        conv_md2rest(coll_changelog_path, coll_changelog_rest_path)
    else:
        logging.warning(
            "Missing collection CHANGELOG.md {0}".format(coll_changelog_path)
        )


def update_collection(args, galaxy, coll_rel):
    """
    Update refs in collection_release.yml.

    Use the latest tag for the ref.  If use_commit_hash is True
    and the latest commit is not tagged, use the commit hash of
    the latest commit for the ref.  Or if the role is specified
    using use_commit_hash_role.
    """
    coll_dir = os.path.join(
        args.dest_path, "ansible_collections", galaxy["namespace"], galaxy["name"]
    )
    if os.path.isdir(coll_dir):
        if args.force:
            shutil.rmtree(coll_dir)
        else:
            raise Exception(
                "collection dest_path {} already exists - remove or use --force".format(
                    coll_dir
                )
            )
    os.makedirs(coll_dir, exist_ok=True)
    collection_readme = os.path.join("lsr_role2collection", "collection_readme.md")
    # major, minor, micro, hash
    versions_updated = [False, False, False, False]
    coll_changelog = ""
    for rolename in args.include:
        if not args.skip_git:
            if args.use_commit_hash and rolename not in args.use_commit_hash_role:
                args.use_commit_hash_role.append(rolename)
            cur_ref = coll_rel[rolename]["ref"]
            tag, cm_hash, tag_is_latest, commit_msgs, prev_tag = get_latest_tag_hash(
                args,
                rolename,
                coll_rel[rolename]["ref"],
                coll_rel[rolename].get("org", args.src_owner),
                coll_rel[rolename].get("repo", rolename),
                rolename in args.use_commit_hash_role,
            )
            if tag or cm_hash:
                if tag_is_latest or rolename not in args.use_commit_hash_role:
                    coll_rel[rolename]["ref"] = tag
                else:
                    coll_rel[rolename]["ref"] = cm_hash
                if not tag_is_latest and not args.no_auto_version:
                    logging.debug(
                        f"role {rolename} tag {tag} is not the latest commit {cm_hash}"
                    )
                check_versions_updated(
                    cur_ref, coll_rel[rolename]["ref"], versions_updated
                )
            # If a version update is detected, retrieve the new changelogs from CHANGELOG.md
            if not args.skip_changelog:
                if comp_versions(cur_ref, coll_rel[rolename]["ref"]) < 0:
                    logging.info(
                        "The role %s is updated. Updating the changelog.", rolename
                    )
                    if prev_tag is None:
                        prev_tag = cur_ref
                    _changelog = get_role_changelog(
                        args, rolename, prev_tag, coll_rel[rolename]["ref"], commit_msgs
                    )
                    coll_changelog = "{}\n{}".format(coll_changelog, _changelog)
        role_to_collection(
            args,
            rolename,
            galaxy["namespace"],
            galaxy["name"],
            collection_readme,
        )
        this_collection = galaxy["namespace"] + "." + galaxy["name"]
        legacy_rqf = "requirements.yml"
        coll_rqf = "collection-requirements.yml"
        for rqf in [legacy_rqf, coll_rqf]:
            req_yml = os.path.join(args.src_path, rolename, "meta", rqf)
            if os.path.isfile(req_yml):
                req_yml_hsh = yaml.safe_load(open(req_yml))
                if isinstance(req_yml_hsh, list):
                    continue  # legacy role format
                if rqf == legacy_rqf:
                    logging.warning(
                        "The role %s is still using %s - please convert to %s instead",
                        rolename,
                        rqf,
                        coll_rqf,
                    )
                for coll in req_yml_hsh.get("collections", []):
                    if isinstance(coll, dict):
                        coll_name = coll["name"]
                        coll_ver = coll.get("version", "*")
                    else:
                        coll_name = coll
                        coll_ver = "*"
                    if coll_name != this_collection:
                        galaxy_deps = galaxy.setdefault("dependencies", {})
                        galaxy_deps[coll_name] = coll_ver
    # Existing changelogs
    orig_cl_file = "lsr_role2collection/COLLECTION_CHANGELOG.md"
    # Collection changelog file path
    coll_changelog_path = os.path.join(coll_dir, "docs", "CHANGELOG.md")
    if not args.no_update:
        if not any(versions_updated):
            logging.info(
                "Nothing in the collection was changed - current collection is up-to-date"
            )
            if not args.skip_changelog:
                shutil.copy(orig_cl_file, coll_changelog_path)
                # Convert CHANGELOG.md to CHANGELOG.rst
                if args.changelog_rst:
                    convert_md2rst(coll_dir)
            create_collection_extra_files(args, coll_dir, galaxy)
            return
        update_galaxy_version(args, galaxy, versions_updated)
        if not args.no_update:
            with open(args.collection_release_yml.name, "w") as crf:
                yaml.safe_dump(coll_rel, crf, sort_keys=False)

    # If to-be-appended changelogs are found, update COLLECTION_CHANGELOG.md
    # and copy it to the collection docs dir.
    if not args.skip_changelog:
        if coll_changelog:
            this_cl_file = (
                "CURRENT_VER_CHANGELOG.md" if args.save_current_changelog else ""
            )
            clhandle, clname = tempfile.mkstemp(suffix=".cl", prefix="collection")
            with os.fdopen(clhandle, "w") as clf:
                # Header
                clf.write("Changelog\n=========\n\n")
                # New changelogs
                new_changelog = "[{}] - {}\n---------------------".format(
                    galaxy["version"], datetime.now().date()
                ) + compact_coll_changelog(coll_changelog)
                clf.write(new_changelog)
                if this_cl_file:
                    with open(this_cl_file, "w") as tcl:
                        tcl.write(new_changelog)
                with open(orig_cl_file, "r") as origclf:
                    _clogs = origclf.read()
                    _clogs = re.sub(
                        "^Changelog\n=========\n", "", _clogs, flags=re.MULTILINE
                    )
                    clf.write(_clogs)
                clf.flush()
                # Overwrite the original changelog with the new one.
                logging.info(
                    f"{orig_cl_file} is updated. Please merge to the master branch"
                )
                shutil.copy(clname, orig_cl_file)
                # Copy the new changelog to the docs dir in collection
                shutil.copy(clname, coll_changelog_path)
        else:
            shutil.copy(orig_cl_file, coll_changelog_path)

        # Convert CHANGELOG.md to CHANGELOG.rst
        if args.changelog_rst:
            convert_md2rst(coll_dir)
    build_collection(args, coll_dir, galaxy)


def find(path, name):
    """Find a file 'name' in or under the directory 'path'."""
    for root, dirs, files in os.walk(path):
        if name in files:
            return os.path.join(root, name)


def process_rpm(args, default_galaxy, coll_rel):
    """
    Extract the contents of rpm.
    If the rpm contains galaxy.yml, use the file.
    Else if MANIFEST.json is found, use the info.
    Otherwise, it fails.
    Generate the collection artifact from the collection part
    of the extracted files.
    """
    workdir = tempfile.mkdtemp(suffix=".lsr", prefix="collection")
    cmdlist0 = ["rpm2cpio", args.rpm]
    p0 = subprocess.Popen(cmdlist0, cwd=workdir, stdout=subprocess.PIPE)
    cmdlist1 = ["cpio", "-id"]
    p1 = subprocess.Popen(cmdlist1, cwd=workdir, stdin=p0.stdout)
    p1.communicate()
    p0.communicate()
    logging.debug(f"{' '.join(cmdlist0)} returned {p0.returncode}")
    logging.debug(f"{' '.join(cmdlist1)} returned {p1.returncode}")
    if p0.returncode != 0:
        raise subprocess.CalledProcessError(
            p0.returncode, cmdlist0, p0.stdout, p0.stderr
        )
    if p1.returncode != 0:
        raise subprocess.CalledProcessError(
            p1.returncode, cmdlist1, p1.stdout, p1.stderr
        )

    tmp_coll = "{}/usr/share/ansible/collections/ansible_collections".format(workdir)
    if not os.path.exists(tmp_coll):
        raise Exception("Failed to extract {} from {}".format(tmp_coll, args.rpm))
    shutil.move(tmp_coll, workdir)
    shutil.rmtree("{}/usr".format(workdir))

    # If exists, use galaxy.yml in rpm
    galaxy_yml = find(workdir, "galaxy.yml")
    if galaxy_yml:
        args.galaxy_yml.close()
        args.galaxy_yml = open(galaxy_yml, "r")
        coll_dir = os.path.dirname(galaxy_yml)
        # override galaxy with the one in rpm
        galaxy = yaml.safe_load(args.galaxy_yml)
    # Otherwise, use the values in MANIFEST.json, if any.
    else:
        manifest_json = find(workdir, "MANIFEST.json")
        if manifest_json:
            with open(manifest_json) as mj:
                mj_contents = json.load(mj)
            galaxy = default_galaxy
            galaxy["namespace"] = mj_contents["collection_info"]["namespace"]
            galaxy["name"] = mj_contents["collection_info"]["name"]
            galaxy["version"] = mj_contents["collection_info"]["version"]
            galaxy["authors"] = mj_contents["collection_info"]["authors"]
            galaxy["description"] = mj_contents["collection_info"]["description"]
            coll_dir = os.path.dirname(manifest_json)
            new_galaxy_yml = "{}/galaxy.yml".format(coll_dir)
            args.galaxy_yml.close()
            args.galaxy_yml = open(new_galaxy_yml, "w")
            yaml.dump(galaxy, args.galaxy_yml)
            os.remove(manifest_json)
            files_json = "{}/FILES.json".format(coll_dir)
            if os.path.exists(files_json):
                os.remove(files_json)
        else:
            raise Exception("No galaxy.yml nor MANIFEST.json in {}".format(args.rpm))

    args.src_path = coll_dir
    build_collection(args, coll_dir)
    shutil.rmtree(workdir)

    return galaxy


def check_collection(args, galaxy):
    coll_file = (
        f"{args.dest_path}/{galaxy['namespace']}-{galaxy['name']}-"
        f"{galaxy['version']}.tar.gz"
    )
    gi_config = "lsr_role2collection/galaxy-importer.cfg"
    if os.path.exists(coll_file) and os.path.exists(gi_config):
        _ = run_cmd(
            [sys.executable, "-m", "galaxy_importer.main", coll_file],
            args.dest_path,
            {"GALAXY_IMPORTER_CONFIG": gi_config},
        )


def publish_collection(args, galaxy):
    coll_file = (
        f"{args.dest_path}/{galaxy['namespace']}-{galaxy['name']}-"
        f"{galaxy['version']}.tar.gz"
    )
    if os.path.exists(coll_file):
        cmd = ["ansible-galaxy", "collection", "publish"]
        if args.no_wait:
            cmd.append("--no-wait")
        else:
            cmd.append("-vv")
        cmd.append(coll_file)
        _ = run_cmd(cmd)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--galaxy-yml",
        type=open,
        default=os.environ.get("GALAXY_YML", "galaxy.yml"),
        help="Path/filename for galaxy.yml",
    )
    parser.add_argument(
        "--collection-release-yml",
        type=open,
        default=os.environ.get("COLLECTION_RELEASE_YML", "collection_release.yml"),
        help="Path/filename for collection_release.yml",
    )
    parser.add_argument(
        "--src-path",
        type=str,
        default=os.environ.get("COLLECTION_SRC_PATH"),
        help="Path to the directory containing the local clone of the role repos",
    )
    parser.add_argument(
        "--src-owner",
        type=str,
        default=os.environ.get("COLLECTION_SRC_OWNER", DEFAULT_GIT_ORG),
        help="Name of the role owner, in legacy role format OWNER.ROLENAME",
    )
    parser.add_argument(
        "--dest-path",
        type=str,
        default=os.environ.get(
            "COLLECTION_DEST_PATH", os.environ.get("HOME") + "/.ansible/collections"
        ),
        help=(
            "Path to collection base directory; collection directory structure "
            "will be created in DIR/ansible_collection/NAMESPACE/COLLECTION_NAME; "
            "collection package file will be created in DIR; "
            "default to ${HOME}/.ansible/collections"
        ),
    )
    parser.add_argument(
        "--force",
        default=False,
        action="store_true",
        help="Remove collection destination dir and file before creating",
    )
    parser.add_argument(
        "--no-auto-version",
        default=False,
        action="store_true",
        help=(
            "The script will try to automatically determine the new collection version based "
            "on the semantic version changes in the versions of the tags in the "
            "role repos.  If you do not want this, add this parameter."
        ),
    )
    parser.add_argument(
        "--new-version",
        default=None,
        type=str,
        help="The new semantic version to use for the collection.",
    )
    parser.add_argument(
        "--no-update",
        default=False,
        action="store_true",
        help=(
            "By default, this script will update collection-release-yml "
            "to the latest tags or commit hashes in the upstream role.  Set "
            "this flag if you want to manually update collection-release-yml "
            "and build the collection from that file."
        ),
    )
    parser.add_argument(
        "--publish",
        default=False,
        action="store_true",
        help="Publish the collection to Galaxy.",
    )
    parser.add_argument(
        "--no-wait",
        default=False,
        action="store_true",
        help="Do not wait for the collection to be published.  Default is to wait.",
    )
    parser.add_argument(
        "--use-commit-hash",
        default=False,
        action="store_true",
        help=(
            "By default, only version tags are used for refs - if True, use latest HEAD"
            " commit hash if not tagged"
        ),
    )
    parser.add_argument(
        "--debug",
        default=False,
        action="store_true",
        help="Turn on debug logging.",
    )
    parser.add_argument(
        "--exclude",
        default=[],
        action="append",
        help="Roles to exclude",
    )
    parser.add_argument(
        "--include",
        default=[],
        action="append",
        help="Roles to include",
    )
    parser.add_argument(
        "--rpm",
        default=None,
        type=str,
        help="Path to the rpm file for the input.",
    )
    parser.add_argument(
        "--skip-git",
        default=False,
        action="store_true",
        help="True when work with local src.",
    )
    parser.add_argument(
        "--skip-check",
        default=False,
        action="store_true",
        help="True when skip check with galaxy-importer.",
    )
    parser.add_argument(
        "--use-commit-hash-role",
        default=[],
        action="append",
        help=(
            "Use the latest commit hash instead of the tag for these roles."
            "  Use this option when you want to use the tag for every role "
            "except the named roles e.g. --use-commit-hash-role sshd "
            "--use-commit-hash-role network will use the tag for every role"
            " except sshd and network, which will use the latest commit hash."
            "  Using --use-commit-hash is the same as using --use-commit-hash-role"
            " and specifying every role."
        ),
    )
    parser.add_argument(
        "--ee-base-image",
        default=os.environ.get(
            "EE_BASE_IMAGE", "quay.io/ansible/ansible-runner:latest"
        ),
        help="Base image for Ansible Execution Environment",
    )
    parser.add_argument(
        "--skip-changelog",
        default=False,
        action="store_true",
        help="If true, do not generate changelog or convert to rst.",
    )
    parser.add_argument(
        "--extra-mapping",
        type=str,
        default="",
        help=(
            "This is a comma delimited list of extra mappings to apply when "
            "converting roles to collection - this replaces the given role "
            "name with collection format with the optional given namespace "
            "and collection as well as the given FQCN with other FQCN."
        ),
    )
    parser.add_argument(
        "--changelog-rst",
        default=False,
        action="store_true",
        help="If true, generate CHANGELOG.rst in collection root directory.",
    )
    parser.add_argument(
        "--save-current-changelog",
        default=False,
        action="store_true",
        help="If true, save the changelog for the current collection version as CURRENT_VER_CHANGELOG.md",
    )
    parser.add_argument(
        "--changelog-sections",
        default=["### New Features", "### Bug Fixes", "### Other Changes"],
        action="append",
        help=(
            "List of the title of changelog sections; Default to "
            "['### New Features', '### Bug Fixes', '### Other Changes']"
        ),
    )
    args = parser.parse_args()

    if args.debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    galaxy = yaml.safe_load(args.galaxy_yml)
    args.galaxy_yml.close()
    coll_rel = yaml.safe_load(args.collection_release_yml)
    args.collection_release_yml.close()
    if not args.include:
        args.include = sorted(list(coll_rel.keys()))
    for item in args.exclude:
        del args.include[item]

    # --rpm rpm_file is given.
    if args.rpm:
        # Check rpm_file exists.
        if not os.path.exists(args.rpm):
            logging.error("{}: No such file given with --rpm".format(args.rpm))
            return
        # If rpm_file is relative path, convert it to absolute path.
        if not os.path.isabs(args.rpm):
            rpm_dir = os.path.abspath(os.getcwd())
            args.rpm = "{0}/{1}".format(rpm_dir, args.rpm)

    workdir = None
    if not args.rpm and not args.src_path:
        workdir = tempfile.mkdtemp(suffix=".lsr", prefix="collection")
        args.src_path = workdir
    try:
        if args.rpm:
            galaxy = process_rpm(args, galaxy, coll_rel)
        else:
            update_collection(args, galaxy, coll_rel)
        if not args.skip_check:
            check_collection(args, galaxy)
        else:
            logging.debug("check_collection is skipped.")
        if args.publish:
            publish_collection(args, galaxy)
        logging.info("Done.")
    finally:
        if workdir:
            shutil.rmtree(workdir)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: (c) 2023, Red Hat, Inc.
# SPDX-License-Identifier: MIT
"""
Modify fingerprint in spec file to be system_role:$rolename

lsr_fingerprint.py -
scans files in the templates dir under ./role["lsrrolename"],
if the file contains a string role["reponame"]:role["rolename"],
replaces it with system_role:role["lsrrolename"].

E.g., in metrics, "performancecopilot:ansible-pcp" is replaced with
"system_role:metrics" in roles/bpftrace/templates/bpftrace.conf.j2.
"""

import sys
from os import listdir, walk
from os.path import join, isfile, isdir

roles = [
    {
        "reponame": "willshersystems",
        "rolename": "ansible-sshd",
        "lsrrolename": "sshd",
    },
    {
        "reponame": "performancecopilot",
        "rolename": "ansible-pcp",
        "lsrrolename": "metrics",
    },
]

changedlist = []
# dirs in the current dir
dirs = [d for d in listdir(".") if isdir(d)]
exit_code = 0
for d in dirs:
    for role in roles:
        if role["lsrrolename"] == d:
            oldpair = "{0}:{1}".format(role["reponame"], role["rolename"])
            newpair = "system_role:{0}".format(role["lsrrolename"])
            for root, subdirs, files in walk(d):
                for subdir in subdirs:
                    if subdir == "templates" or subdir == "meta" or subdir == "tests":
                        dirpath = join(root, subdir)
                        tmpls = [
                            f for f in listdir(dirpath) if isfile(join(dirpath, f))
                        ]
                        for tmpl in tmpls:
                            tmplpath = join(dirpath, tmpl)
                            with open(tmplpath) as ifp:
                                lines = ifp.read()
                            if oldpair in lines:
                                with open(tmplpath) as ifp:
                                    lines = ifp.readlines()
                                newlines = []
                                changed = {}
                                for lineno, line in enumerate(lines):
                                    newline = line.replace(oldpair, newpair)
                                    if newline != line:
                                        if len(changed) > 0:
                                            dup = True
                                        else:
                                            dup = False
                                        changed = {
                                            "path": tmplpath,
                                            "linenumber": lineno,
                                            "oldline": line.strip(),
                                            "newline": newline.strip(),
                                            "duplicate": dup,
                                            "test": subdir == "tests",
                                        }
                                        changedlist.append(changed)
                                    newlines.append(newline)
                                with open(tmplpath, "w") as ofp:
                                    ofp.writelines(newlines)
            print(
                "{0} role done. Made the following changes:".format(role["lsrrolename"])
            )
            for changed in changedlist:
                # allow duplicates in tests
                dup = changed["duplicate"] and not changed["test"]
                print(
                    "{0}{1}:{2}".format(
                        "DUPLICATED - " if dup else "",
                        changed["path"],
                        changed["linenumber"],
                    )
                )
                if dup:
                    print("  questionable line: {0}".format(changed["oldline"]))
                    exit_code = 1
                else:
                    print("  old line: {0}".format(changed["oldline"]))
                    print("  new line: {0}".format(changed["newline"]))
sys.exit(exit_code)



#!/usr/bin/python3

import sys
import io
import re
import yaml


COLLECTION_RELEASE = "collection_release.yml"
SPECFILE_IN = sys.argv[1]
SPECFILE_OUT = sys.argv[2]

BEGIN_SOURCES_MARKER = "# BEGIN AUTOGENERATED SOURCES"
END_SOURCES_MARKER = "# END AUTOGENERATED SOURCES"

BEGIN_SETUP_MARKER = "# BEGIN AUTOGENERATED SETUP"
END_SETUP_MARKER = "# END AUTOGENERATED SETUP"

BEGIN_SETUP_TMPL = "%setup -q "
END_SETUP_TMPL = " -n %{getarchivedir 0}"


class Replacement:
    """Class describing a replacement of lines between markers in a file"""

    def __init__(self, begin_marker, end_marker, generate):
        self.begin_marker = begin_marker
        self.end_marker = end_marker
        self.generate = generate
        self.throwaway = False
        self.marker_seen = False

    def process_line(self, line):
        # did we change the input line? If not, we are merely returning it.
        changed = False
        of = io.StringIO()
        if not self.throwaway:
            of.write(line)
            if line.strip() == self.begin_marker:
                if self.marker_seen:
                    sys.exit("Duplicate replacement section")
                changed = True
                self.generate(of)
                self.throwaway = True
                self.marker_seen = True
            elif line.strip() == self.end_marker:
                sys.exit("End marker '{}' without begin marker".format(self.end_marker))
        else:
            if line.strip() == self.end_marker:
                self.throwaway = False
                of.write(line)
            elif line.strip() == self.begin_marker:
                sys.exit("Duplicate begin marker: {}".format(self.begin_marker))
            else:
                # we discarded the input line
                changed = True

        return (changed, of.getvalue())

    def inside(self):
        """Are we currently inside the replacement block?"""
        return self.throwaway

    def after(self):
        """Have we been inside the replacement block?"""
        return self.marker_seen and not self.throwaway


def ref_is_commit(ref):
    if re.match(r"[0-9a-f]{5,40}$", ref):
        # is a SHA-1 commit id
        return True
    elif re.search(r"[0-9]+[.][0-9]+[.][0-9]+", ref):
        # is a tag
        # search() and not match() because there could be something in front
        # of the semantic version number, like literal "v" - we have tags looking
        # both like "v1.2.0" and like "1.2.0"
        return False
    else:
        sys.exit(f"unknown format of reference '{ref}': neither commit nor semver tag")


def generate_sources(output):
    k = list(rolesbysourcenum.keys())
    k.sort()

    for i in k:
        r = rolesbysourcenum[i]
        org = roles[r].get("org")
        repo = roles[r].get("repo")
        ref = roles[r]["ref"]
        if org:
            # If we support other hosting service than Github, the format
            # of collection_release.yml will have to change a bit.
            output.write(f"%global forgeorg{i} https://github.com/{org}\n")
        if repo:
            output.write(f"%global repo{i} {repo}\n")
        output.write(f"%global rolename{i} {r}\n")
        if ref_is_commit(ref):
            output.write(f"%defcommit {i} {ref}\n\n")
        else:
            output.write(f"%deftag {i} {ref}\n\n")
    for i in k:
        output.write(f"Source{i}: %{{archiveurl{i}}}\n")


def generate_setup(output):
    k = list(rolesbysourcenum.keys())
    k.sort()

    output.write(
        BEGIN_SETUP_TMPL + " ".join(f"-a{i}" for i in k) + END_SETUP_TMPL + "\n"
    )


with open(COLLECTION_RELEASE, "r") as cf:
    roles = yaml.safe_load(cf)
    for rolename, roleinfo in roles.items():
        print(roleinfo)
    rolesbysourcenum = {
        roleinfo["sourcenum"]: rolename for rolename, roleinfo in roles.items()
    }


sources_replacement = Replacement(
    BEGIN_SOURCES_MARKER, END_SOURCES_MARKER, generate_sources
)
setup_replacement = Replacement(BEGIN_SETUP_MARKER, END_SETUP_MARKER, generate_setup)

f = open(SPECFILE_IN, "r")
of = open(SPECFILE_OUT, "w")

for line in f:
    changed_sources, new_line = sources_replacement.process_line(line)
    changed_setup, new_new_line = setup_replacement.process_line(new_line)
    if changed_sources and changed_setup:
        sys.exit(
            "Begin marker {} in the output of sources replacement".format(
                setup_replacement.begin_marker
            )
        )
    of.write(new_new_line)

for i in (sources_replacement, setup_replacement):
    if i.inside():
        sys.exit("Missing end marker '{}'".format(i.end_marker))
    elif not i.after():
        sys.exit("Missing begin marker '{}'".format(i.begin_marker))



#!/usr/bin/python3

import yaml


def get_collection_version():
    with open("galaxy.yml", "r") as gf:
        galaxy = yaml.safe_load(gf)
        return galaxy["version"]


if __name__ == "__main__":
    print(get_collection_version())
